
/**************************************************************************
 *
 *  $Id: mbg_ntp_auth.c 1.4 2019/09/04 07:38:26 martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Crypto functions useful for NTP implementations, stolen from
 *    original NTP source code.
 *
 * -----------------------------------------------------------------------
 *  $Log: mbg_ntp_auth.c $
 *  Revision 1.4  2019/09/04 07:38:26  martin
 *  Openssl v1.1.1 support by Marius, but modified to avoid duplicate code.
 *  ereallocz() doesn't exit on error with an arbitrary exit code. Instead,
 *  it returns NULL, so the calling application can/has to handle the error,
 *  e.g. exit with an appropriate code.
 *  mbg_ntp_auth_init() now returns MBG_SUCCESS, or an appropriate
 *  Meinberg error code.
 *  Removed trailing spaces.
 *  Revision 1.3  2018/08/01 14:37:11  martin
 *  Fixed build on Windows.
 *  Revision 1.2  2014/09/24 08:55:12Z  martin
 *  Full support of NTP symmetric key authentication.
 *  Export only Meinberg-specific functions which call
 *  the NTP functions internally.
 *  Modifications to make the key cache thread-safe.
 *  Locking will be required in addition if keys are added,
 *  deleted, or made untrusted at run time.
 *  Revision 1.1  2014/09/10 15:37:57  martin
 *  Initial revision.
 *
 **************************************************************************/

#define _MBG_NTP_AUTH
  #include <mbg_ntp_auth.h>
#undef _MBG_NTP_AUTH

#include <mbgerror.h>

#include <stdlib.h>
#include <string.h>

#include <stddef.h>  // offsetof()
#include <math.h>    // log10()

#if !defined( MBG_TGT_WIN32 )
  #include <syslog.h>  // log levels
  #include <arpa/inet.h>
#else
  #define LOG_ERR      0  // TODO
  #define LOG_WARNING  1
#endif

#include <ctype.h>   // toupper()

#if defined( MBG_TGT_WIN32 )
  #define strdup _strdup
#endif


#if NTP_COMPAT
  #define TRUE 1
  #define FALSE 0

  #if defined( MBG_TGT_WIN32 )
    #define zero_mem(p, s)  memset(p, 0, s)
  #else
    #define zero_mem(p, s)  bzero(p, s)
    #define max(n1, n2)  ( ( (n1) > (n2) ) ? (n1) : (n2) )
    #define min(n1, n2)  ( ( (n1) < (n2) ) ? (n1) : (n2) )
  #endif



  #define DEBUG_ENSURE( _v )  do {} while(0)
  static u_long   current_time;           /* seconds since startup */

  #ifdef DEBUG
    static int debug = DEBUG;
  #endif

/*
 * macro for debugging output - cut down on #ifdef pollution.
 *
 * DPRINTF() is for use by ntpd only, and compiles away to nothing
 * without DEBUG (configure --disable-debugging).
 *
 * TRACE() is similar for libntp and utilities, which retain full
 * debug capability even when compiled without DEBUG.
 *
 * The calling convention is not attractive:
 *     DPRINTF(debuglevel, (fmt, ...));
 *     DPRINTF(2, ("shows #ifdef DEBUG and if debug >= %d\n", 2));
 */
#if 0 && defined( DEBUG )  //##++++++
# define DPRINTF(lvl, arg)                              \
        do {                                            \
                if (debug >= (lvl))                     \
                        mprintf arg;                    \
        } while (0)
#else
# define DPRINTF(lvl, arg)      do {} while (0)
#endif

  #define msyslog mbglog
  extern void mbglog( int lvl, const char *fmt, ... );
#endif  // NTP_COMPAT




#if 1 // SLISTS from ntp_lists.h

#ifdef DEBUG
# define NTP_DEBUG_LISTS_H
#endif

/*
 * If list debugging is not enabled, save a little time by not clearing
 * an entry's link pointer when it is unlinked, as the stale pointer
 * is harmless as long as it is ignored when the entry is not in a
 * list.
 */
#ifndef NTP_DEBUG_LISTS_H
#define MAYBE_Z_LISTS(p)        do { } while (FALSE)
#else
#define MAYBE_Z_LISTS(p)        (p) = NULL
#endif

#define LINK_SLIST(listhead, pentry, nextlink)                  \
do {                                                            \
        (pentry)->nextlink = (listhead);                        \
        (listhead) = (pentry);                                  \
} while (FALSE)

#define LINK_TAIL_SLIST(listhead, pentry, nextlink, entrytype)  \
do {                                                            \
        entrytype **pptail;                                     \
                                                                \
        pptail = &(listhead);                                   \
        while (*pptail != NULL)                                 \
                pptail = &((*pptail)->nextlink);                \
                                                                \
        (pentry)->nextlink = NULL;                              \
        *pptail = (pentry);                                     \
} while (FALSE)

#define LINK_SORT_SLIST_CURRENT()       (*ppentry)
#define L_S_S_CUR()                     LINK_SORT_SLIST_CURRENT()

#define LINK_SORT_SLIST(listhead, pentry, beforecur, nextlink,  \
                        entrytype)                              \
do {                                                            \
        entrytype **ppentry;                                    \
                                                                \
        ppentry = &(listhead);                                  \
        while (TRUE) {                                          \
                if (NULL == *ppentry || (beforecur)) {          \
                        (pentry)->nextlink = *ppentry;          \
                        *ppentry = (pentry);                    \
                        break;                                  \
                }                                               \
                ppentry = &((*ppentry)->nextlink);              \
                if (NULL == *ppentry) {                         \
                        (pentry)->nextlink = NULL;              \
                        *ppentry = (pentry);                    \
                        break;                                  \
                }                                               \
        }                                                       \
} while (FALSE)

#define UNLINK_HEAD_SLIST(punlinked, listhead, nextlink)        \
do {                                                            \
        (punlinked) = (listhead);                               \
        if (NULL != (punlinked)) {                              \
                (listhead) = (punlinked)->nextlink;             \
                MAYBE_Z_LISTS((punlinked)->nextlink);           \
        }                                                       \
} while (FALSE)

#define UNLINK_EXPR_SLIST_CURRENT()     (*ppentry)
#define U_E_S_CUR()                     UNLINK_EXPR_SLIST_CURRENT()

#define UNLINK_EXPR_SLIST(punlinked, listhead, expr, nextlink,  \
                          entrytype)                            \
do {                                                            \
        entrytype **ppentry;                                    \
                                                                \
        ppentry = &(listhead);                                  \
                                                                \
        while (!(expr))                                         \
                if (*ppentry != NULL &&                         \
                    (*ppentry)->nextlink != NULL) {             \
                        ppentry = &((*ppentry)->nextlink);      \
                } else {                                        \
                        ppentry = NULL;                         \
                        break;                                  \
                }                                               \
                                                                \
        if (ppentry != NULL) {                                  \
                (punlinked) = *ppentry;                         \
                *ppentry = (punlinked)->nextlink;               \
                MAYBE_Z_LISTS((punlinked)->nextlink);           \
        } else {                                                \
                (punlinked) = NULL;                             \
        }                                                       \
} while (FALSE)

#define UNLINK_SLIST(punlinked, listhead, ptounlink, nextlink,  \
                     entrytype)                                 \
        UNLINK_EXPR_SLIST(punlinked, listhead, (ptounlink) ==   \
            U_E_S_CUR(), nextlink, entrytype)

#define CHECK_SLIST(listhead, nextlink, entrytype)              \
do {                                                            \
        entrytype *pentry;                                      \
                                                                \
        for (pentry = (listhead);                               \
             pentry != NULL;                                    \
             pentry = pentry->nextlink){                        \
                NTP_INSIST(pentry != pentry->nextlink);         \
                NTP_INSIST((listhead) != pentry->nextlink);     \
        }                                                       \
} while (FALSE)

#endif  // SLISTS from ntp_lists.h





#if 1  // from ntp_lists.h

/*
 * DLIST
 */
#define DECL_DLIST_LINK(entrytype, link)                        \
struct {                                                        \
        entrytype *     b;                                      \
        entrytype *     f;                                      \
} link

#define INIT_DLIST(listhead, link)                              \
do {                                                            \
        (listhead).link.f = &(listhead);                        \
        (listhead).link.b = &(listhead);                        \
} while (FALSE)

#define HEAD_DLIST(listhead, link)                              \
        (                                                       \
                (&(listhead) != (listhead).link.f)              \
                    ? (listhead).link.f                         \
                    : NULL                                      \
        )

#define TAIL_DLIST(listhead, link)                              \
        (                                                       \
                (&(listhead) != (listhead).link.b)              \
                    ? (listhead).link.b                         \
                    : NULL                                      \
        )

#define NEXT_DLIST(listhead, entry, link)                       \
        (                                                       \
                (&(listhead) != (entry)->link.f)                \
                    ? (entry)->link.f                           \
                    : NULL                                      \
        )

#define PREV_DLIST(listhead, entry, link)                       \
        (                                                       \
                (&(listhead) != (entry)->link.b)                \
                    ? (entry)->link.b                           \
                    : NULL                                      \
        )

#define LINK_DLIST(listhead, pentry, link)                      \
do {                                                            \
        (pentry)->link.f = (listhead).link.f;                   \
        (pentry)->link.b = &(listhead);                         \
        (listhead).link.f->link.b = (pentry);                   \
        (listhead).link.f = (pentry);                           \
} while (FALSE)

#define LINK_TAIL_DLIST(listhead, pentry, link)                 \
do {                                                            \
        (pentry)->link.b = (listhead).link.b;                   \
        (pentry)->link.f = &(listhead);                         \
        (listhead).link.b->link.f = (pentry);                   \
        (listhead).link.b = (pentry);                           \
} while (FALSE)

#define UNLINK_DLIST(ptounlink, link)                           \
do {                                                            \
        (ptounlink)->link.b->link.f = (ptounlink)->link.f;      \
        (ptounlink)->link.f->link.b = (ptounlink)->link.b;      \
        MAYBE_Z_LISTS((ptounlink)->link.b);                     \
        MAYBE_Z_LISTS((ptounlink)->link.f);                     \
} while (FALSE)

#define ITER_DLIST_BEGIN(listhead, iter, link, entrytype)       \
{                                                               \
        entrytype *i_dl_nextiter;                               \
                                                                \
        for ((iter) = (listhead).link.f;                        \
             (iter) != &(listhead)                              \
             && ((i_dl_nextiter = (iter)->link.f), TRUE);       \
             (iter) = i_dl_nextiter) {
#define ITER_DLIST_END()                                        \
        }                                                       \
}

#define REV_ITER_DLIST_BEGIN(listhead, iter, link, entrytype)   \
{                                                               \
        entrytype *i_dl_nextiter;                               \
                                                                \
        for ((iter) = (listhead).link.b;                        \
             (iter) != &(listhead)                              \
             && ((i_dl_nextiter = (iter)->link.b), TRUE);       \
             (iter) = i_dl_nextiter) {
#define REV_ITER_DLIST_END()                                    \
        }                                                       \
}

#endif  // ntp_lists.h




#if 1  // from ntp_malloc.h

#ifdef EREALLOC_IMPL
# define EREALLOC_CALLSITE      /* preserve __FILE__ and __LINE__ */
#else
# define EREALLOC_IMPL(ptr, newsz, filenm, loc) \
         realloc(ptr, (newsz))
#endif

#endif // from ntp_malloc.h




#if 1  // from ntp_stdlib.h

/* emalloc.c */
#ifndef EREALLOC_CALLSITE       /* ntp_malloc.h defines */
static /*extern*/  void *  ereallocz       (void *, size_t, size_t, int);
#define erealloczsite(p, n, o, z, f, l) ereallocz(p, n, o, (z))
static /*extern*/  void *  emalloc         (size_t);
#define emalloc_zero(c)         ereallocz(NULL, (c), 0, TRUE)
#define erealloc(p, c)          ereallocz(p, (c), 0, FALSE)
#define erealloc_zero(p, n, o)  ereallocz(p, n, (o), TRUE)
extern  char *  estrdup_impl    (const char *);
#define estrdup(s)              estrdup_impl(s)
#else
extern  void *  ereallocz       (void *, size_t, size_t, int,
                                 const char *, int);
#define erealloczsite           ereallocz
#define emalloc(c)              ereallocz(NULL, (c), 0, FALSE, \
                                          __FILE__, __LINE__)
#define emalloc_zero(c)         ereallocz(NULL, (c), 0, TRUE, \
                                          __FILE__, __LINE__)
#define erealloc(p, c)          ereallocz(p, (c), 0, FALSE, \
                                          __FILE__, __LINE__)
#define erealloc_zero(p, n, o)  ereallocz(p, n, (o), TRUE, \
                                          __FILE__, __LINE__)
extern  char *  estrdup_impl    (const char *, const char *, int);
#define estrdup(s) estrdup_impl((s), __FILE__, __LINE__)
#endif

/* ssl_init.c */
#ifdef OPENSSL
static /*extern*/  void    ssl_init                (void);
// static /*extern*/  void    ssl_check_version       (void);
static /*extern*/  int     ssl_init_done;
#define INIT_SSL()                              \
        do {                                    \
                if (!ssl_init_done)             \
                        ssl_init();             \
        } while (0)
#else   /* !OPENSSL follows */
#define INIT_SSL()              do {} while (0)
#endif

#endif  // from ntp_stdlib.h




#if 1  // from ssl_init.c

#ifdef OPENSSL

static
void    atexit_ssl_cleanup(void);

static
int ssl_init_done;

static
void
ssl_init(void)
{
//##+++++        init_lib();

        if (ssl_init_done)
                return;

        ERR_load_crypto_strings();
        //##+++++ ERR_load_BIO_strings(); // TODO: NTP doesn't use this. Do we need it?
        OpenSSL_add_all_algorithms();
        atexit(&atexit_ssl_cleanup);

        ssl_init_done = TRUE;
}


static
void
atexit_ssl_cleanup(void)
{
        if (!ssl_init_done)
                return;

        ssl_init_done = FALSE;
        EVP_cleanup();
        ERR_free_strings();
}



#if 0

static
void
ssl_check_version(void)
{
        if ((SSLeay() ^ OPENSSL_VERSION_NUMBER) & ~0xff0L) {
                msyslog(LOG_WARNING,
                    "OpenSSL version mismatch. Built against %lx, you have %lx",
                    (u_long)OPENSSL_VERSION_NUMBER, SSLeay());
                fprintf(stderr,
                    "OpenSSL version mismatch. Built against %lx, you have %lx\n",
                    (u_long)OPENSSL_VERSION_NUMBER, SSLeay());
        }

        INIT_SSL();
}

#endif // 0

#endif  /* OPENSSL */


/*
 * keytype_name         returns OpenSSL short name for digest by NID.
 *
 * Used by ntpq and ntpdc keytype()
 */
static
const char *
keytype_name(
        int nid
        )
{
        static const char unknown_type[] = "(unknown key type)";
        const char *name;

#ifdef OPENSSL
        INIT_SSL();
        name = OBJ_nid2sn(nid);
        if (NULL == name)
                name = unknown_type;
#else   /* !OPENSSL follows */
        if (NID_md5 == nid)
                name = "MD5";
        else
                name = unknown_type;
#endif
        return name;
}



static /*HDR*/
EVP_MD_CTX *get_md_ctx( void )
{
  EVP_MD_CTX *ctx;

  #if OPENSSL_VERSION_NUMBER >= 0x010101000
    ctx = EVP_MD_CTX_new();
  #else
    ctx = EVP_MD_CTX_create();
  #endif

  if ( ctx == NULL )
    msyslog( LOG_ERR, "MAC encrypt: ctx init failed");

  return ctx;

}  // get_md_ctx



/*
 * keytype_from_text    returns OpenSSL NID for digest by name, and
 *                      optionally the associated digest length.
 *
 * Used by ntpd authreadkeys(), ntpq and ntpdc keytype()
 */
static
int
keytype_from_text(
        const char *text,
        size_t *pdigest_len
        )
{
        int             key_type;
        u_int           digest_len;
#ifdef OPENSSL
        const u_long    max_digest_len = MAX_MAC_LEN - sizeof(keyid_t);
        u_char          digest[EVP_MAX_MD_SIZE];
        char *          upcased;
        char *          pch;
        EVP_MD_CTX *    ctx;

        /*
         * OpenSSL digest short names are capitalized, so uppercase the
         * digest name before passing to OBJ_sn2nid().  If it is not
         * recognized but begins with 'M' use NID_md5 to be consistent
         * with past behavior.
         */
        INIT_SSL();

        ctx = get_md_ctx();

        if ( ctx == NULL )
          return 0;

      #if 1  // NOTE: modification by martin
        upcased = strdup( text );
        if ( upcased )
        {
          for (pch = upcased; '\0' != *pch; pch++)
            *pch = (char)toupper(*pch);
          key_type = OBJ_sn2nid( upcased );
          free( upcased );
        }
        else
          key_type = 0;
      #else
        LIB_GETBUF(upcased);
        strlcpy(upcased, text, LIB_BUFLENGTH);
        for (pch = upcased; '\0' != *pch; pch++)
                *pch = (char)toupper(*pch);
        key_type = OBJ_sn2nid(upcased);
      #endif
#else
        key_type = 0;
#endif

        if (!key_type && 'm' == tolower(text[0]))
                key_type = NID_md5;

        if (!key_type)
                return 0;

        if (NULL != pdigest_len) {
#ifdef OPENSSL
                EVP_DigestInit(ctx, EVP_get_digestbynid(key_type));
                EVP_DigestFinal(ctx, digest, &digest_len);
                if (digest_len > max_digest_len) {
                        fprintf(stderr,
                                "key type %s %u octet digests are too big, max %lu\n",
                                keytype_name(key_type), digest_len,
                                max_digest_len);
                        msyslog(LOG_ERR,
                                "key type %s %u octet digests are too big, max %lu",
                                keytype_name(key_type), digest_len,
                                max_digest_len);
                        return 0;
                }
#else
                digest_len = 16;
#endif
                *pdigest_len = digest_len;
        }

        return key_type;
}

#endif  // from ssl_init.c



/*
 * Structure to store keys in in the hash table.
 */
typedef struct savekey symkey;

struct savekey {
        symkey *        hlink;          /* next in hash bucket */
        DECL_DLIST_LINK(symkey, llink); /* for overall & free lists */
        // TODO: the next fields except 'lifetime' are also
        // used by the key cache.
        u_char *        secret;         /* shared secret */
        u_long          lifetime;       /* remaining lifetime */
        keyid_t         keyid;          /* key identifier */
        u_short         type;           /* OpenSSL digest NID */
        u_short         secretsize;     /* secret octets */
        u_short         flags;          /* KEY_ flags that wave */
};

/* define the payload region of symkey beyond the list pointers */
#define symkey_payload  secret

#define KEY_TRUSTED     0x001   /* this key is trusted */

#ifdef DEBUG
typedef struct symkey_alloc_tag symkey_alloc;

struct symkey_alloc_tag {
        symkey_alloc *  link;
        void *          mem;            /* enable free() atexit */
};

symkey_alloc *  authallocs;
#endif  /* DEBUG */

static __mbg_inline u_short   auth_log2(double x);
static void             auth_resize_hashtable(void);
static void             allocsymkey(symkey **, keyid_t, u_short,
                                    u_short, u_long, u_short, u_char *);
static void             freesymkey(symkey *, symkey **);
#ifdef DEBUG_KEYS
static void             free_auth_mem(KEY_CACHE *);
#endif

symkey  key_listhead;           /* list of all in-use keys */;
/*
 * The hash table. This is indexed by the low order bits of the
 * keyid. We make this fairly big for potentially busy servers.
 */
#define DEF_AUTHHASHSIZE        64
//#define       HASHMASK        ((HASHSIZE)-1)
#define KEYHASH(keyid)  ((keyid) & authhashmask)

int     authhashdisabled;
u_short authhashbuckets = DEF_AUTHHASHSIZE;
u_short authhashmask = DEF_AUTHHASHSIZE - 1;
symkey **key_hash;

u_long authkeynotfound;         /* keys not found */
u_long authkeylookups;          /* calls to lookup keys */
u_long authnumkeys;             /* number of active keys */
u_long authkeyexpired;          /* key lifetime expirations */
u_long authkeyuncached;         /* cache misses */
u_long authnokey;               /* calls to encrypt with no key */
u_long authencryptions;         /* calls to encrypt */
u_long authdecryptions;         /* calls to decrypt */

/*
 * Storage for free symkey structures.  We malloc() such things but
 * never free them.
 */
symkey *authfreekeys;
int authnumfreekeys;

#define MEMINC  16              /* number of new free ones to get */

#if 1  // from a_md5encrypt.c

/*
 * MD5authencrypt - generate message digest
 *
 * Returns length of MAC including key ID and digest.
 */
static
int
MD5authencrypt(
        KEY_CACHE *p_cache,     /* TODO: check cache and parameter usage */
        int     type,           /* hash algorithm */
        u_char  *key,           /* key pointer */
        u_int32 *pkt,           /* packet pointer */
        int     length          /* packet length */
        )
{
        u_char  digest[EVP_MAX_MD_SIZE];
        u_int   len;
        EVP_MD_CTX * ctx;

        /*
         * Compute digest of key concatenated with packet. Note: the
         * key type and digest type have been verified when the key
         * was creaded.
         */
        INIT_SSL();

        ctx = get_md_ctx();

        if ( ctx == NULL )
          return 0;

#if defined(OPENSSL) && OPENSSL_VERSION_NUMBER >= 0x0090700fL
        if (!EVP_DigestInit(ctx, EVP_get_digestbynid(type))) {
                msyslog(LOG_ERR,
                    "MAC encrypt: digest init failed");
                return (0);
        }
#else
        EVP_DigestInit(ctx, EVP_get_digestbynid(type));
#endif
        EVP_DigestUpdate(ctx, key, p_cache->secretsize);
        EVP_DigestUpdate(ctx, (u_char *)pkt, (u_int)length);
        EVP_DigestFinal(ctx, digest, &len);
        memmove((u_char *)pkt + length + 4, digest, len);
        return (len + 4);
}


/*
 * MD5authdecrypt - verify MD5 message authenticator
 *
 * Returns one if digest valid, zero if invalid.
 */
static
int
MD5authdecrypt(
        KEY_CACHE *p_cache,     /* TODO: check cache and parameter usage */
        int     type,           /* hash algorithm */
        u_char  *key,           /* key pointer */
        u_int32 *pkt,           /* packet pointer */
        int     length,         /* packet length */
        int     size            /* MAC size */
        )
{
        u_char  digest[EVP_MAX_MD_SIZE];
        u_int   len;
        EVP_MD_CTX * ctx;

        /*
         * Compute digest of key concatenated with packet. Note: the
         * key type and digest type have been verified when the key
         * was created.
         */
        INIT_SSL();

        ctx = get_md_ctx();

        if ( ctx == NULL )
          return 0;

#if defined(OPENSSL) && OPENSSL_VERSION_NUMBER >= 0x0090700fL
        if (!EVP_DigestInit(ctx, EVP_get_digestbynid(type))) {
                msyslog(LOG_ERR,
                    "MAC decrypt: digest init failed");
                return (0);
        }
#else
        EVP_DigestInit(ctx, EVP_get_digestbynid(type));
#endif
        EVP_DigestUpdate(ctx, key, p_cache->secretsize);
        EVP_DigestUpdate(ctx, (u_char *)pkt, (u_int)length);
        EVP_DigestFinal(ctx, digest, &len);
        if ((u_int)size != len + 4) {
                msyslog(LOG_ERR,
                    "MAC decrypt: MAC length error");
                return (0);
        }

        #if DEBUG_MBG_NTP_AUTH
        {
          NTP_PACKET *p = (NTP_PACKET *) pkt;
          unsigned char *cp;
          int i;

          #define N_NTP_MAC_BYTES  16

          printf( "using MD5 key ID:   %08X, len = %i\n", htonl( p->mac.key_id ), len );

          printf( "packet MD5 key hash: " );

          for ( i = 0, cp = (unsigned char *)pkt + length + 4; i < N_NTP_MAC_BYTES; i++, cp++ )
            printf( "%02X", *cp );

          printf( "\n" );

          printf( "digest MD5 key hash: " );

          for ( i = 0, cp = digest; i < N_NTP_MAC_BYTES; i++, cp++ )
            printf( "%02X", *cp );

          printf( "\n" );
        }
        #endif

        return !memcmp(digest, (char *)pkt + length + 4, len);
}



#if 0  //##++++++++++++++++++++

/*
 * Calculate the reference id from the address. If it is an IPv4
 * address, use it as is. If it is an IPv6 address, do a md5 on
 * it and use the bottom 4 bytes.
 * The result is in network byte order.
 */
u_int32
addr2refid(sockaddr_u *addr)
{
        u_char          digest[20];
        u_int32         addr_refid;
        EVP_MD_CTX *    ctx;
        u_int           len;

        if (IS_IPV4(addr))
                return (NSRCADR(addr));

        INIT_SSL();

        ctx = get_md_ctx();

        if ctx == NULL
          return 0;


#if defined(OPENSSL) && OPENSSL_VERSION_NUMBER >= 0x0090700fL
  #ifdef EVP_MD_CTX_FLAG_NON_FIPS_ALLOW
        /* MD5 is not used as a crypto hash here. */
        EVP_MD_CTX_set_flags(ctx, EVP_MD_CTX_FLAG_NON_FIPS_ALLOW);
  #endif
        if (!EVP_DigestInit_ex(ctx, EVP_md5(), NULL)) {
                msyslog(LOG_ERR,
                    "MD5 init failed");
                // TODO We should not exit with an arbitrary code here.
                // Instead, we should return an error to the caller
                // and let the application determine an appropriate
                // exit code.
                exit(1);
        }
#else
        EVP_DigestInit(ctx, EVP_md5());
#endif

        EVP_DigestUpdate(ctx, (u_char *)PSOCK_ADDR6(addr),
            sizeof(struct in6_addr));
        EVP_DigestFinal(ctx, digest, &len);
        memcpy(&addr_refid, digest, sizeof(addr_refid));
        return (addr_refid);
}

#endif   // 0 //##++++++

#endif   // from a_md5encrypt.c





#if 1  // from emalloc.c

/*
 * When using the debug MS CRT allocator, each allocation stores the
 * callsite __FILE__ and __LINE__, which is then displayed at process
 * termination, to track down leaks.  We don't want all of our
 * allocations to show up as coming from emalloc.c, so we preserve the
 * original callsite's source file and line using macros which pass
 * __FILE__ and __LINE__ as parameters to these routines.
 * Other debug malloc implementations can be used by defining
 * EREALLOC_IMPL() as ports/winnt/include/config.h does.
 */

static
void *
ereallocz(
        void *  ptr,
        size_t  newsz,
        size_t  priorsz,
        int     zero_init
#ifdef EREALLOC_CALLSITE                /* ntp_malloc.h */
                         ,
        const char *    file,
        int             line
#endif
        )
{
        char *  mem;
        size_t  allocsz;

        if (0 == newsz)
                allocsz = 1;
        else
                allocsz = newsz;

        mem = EREALLOC_IMPL(ptr, allocsz, file, line);
        if (NULL == mem) {
//##+++++                msyslog_term = TRUE;
#ifndef EREALLOC_CALLSITE
                msyslog(LOG_ERR, "fatal out of memory (%lu bytes)",
                        (u_long)newsz);
#else
                msyslog(LOG_ERR,
                        "fatal out of memory %s line %d (%lu bytes)",
                        file, line, (u_long)newsz);
#endif
                return NULL;
        }

        if (zero_init && newsz > priorsz)
                zero_mem(mem + priorsz, newsz - priorsz);

        return mem;
}

//##++++ [...]

#ifndef EREALLOC_CALLSITE
static
void *
emalloc(size_t newsz)
{
        return ereallocz(NULL, newsz, 0, FALSE);
}
#endif


#endif  // from emalloc.c



/*
 * init_auth - initialize internal data
 */
static
int
init_auth(void)
{
        size_t newalloc;

        /*
         * Initialize hash table and free list
         */
        newalloc = authhashbuckets * sizeof(key_hash[0]);

        key_hash = erealloc(key_hash, newalloc);

        if ( key_hash == NULL )
        {
          errno = ENOMEM;
          return -1;
        }

        memset(key_hash, '\0', newalloc);

        INIT_DLIST(key_listhead, llink);

#ifdef DEBUG_KEYS
        atexit(&free_auth_mem);
#endif

  return 0;  // success
}


/*
 * free_auth_mem - assist in leak detection by freeing all dynamic
 *                 allocations from this module.
 */
#ifdef DEBUG_KEYS
static void
free_auth_mem( KEY_CACHE *p_cache )
{
        symkey *        sk;
        symkey_alloc *  alloc;
        symkey_alloc *  next_alloc;

        while (NULL != (sk = HEAD_DLIST(key_listhead, llink))) {
                freesymkey(sk, &key_hash[KEYHASH(sk->keyid)]);
        }
        free(key_hash);
        key_hash = NULL;
        p_cache->keyid = 0;
        p_cache->flags = 0;
        for (alloc = authallocs; alloc != NULL; alloc = next_alloc) {
                next_alloc = alloc->link;
                free(alloc->mem);
        }
        authfreekeys = NULL;
        authnumfreekeys = 0;
}
#endif  /* DEBUG */


/*
 * auth_moremem - get some more free key structures
 */
static
void
auth_moremem(
        int     keycount
        )
{
        symkey *        sk;
        int             i;
#ifdef DEBUG
        void *          base;
        symkey_alloc *  allocrec;
# define MOREMEM_EXTRA_ALLOC    (sizeof(*allocrec))
#else
# define MOREMEM_EXTRA_ALLOC    (0)
#endif

        i = (keycount > 0)
                ? keycount
                : MEMINC;
        sk = emalloc_zero(i * sizeof(*sk) + MOREMEM_EXTRA_ALLOC);
#ifdef DEBUG
        base = sk;
#endif
        authnumfreekeys += i;

        for (; i > 0; i--, sk++) {
                LINK_SLIST(authfreekeys, sk, llink.f);
        }

#ifdef DEBUG
        allocrec = (void *)sk;
        allocrec->mem = base;
        LINK_SLIST(authallocs, allocrec, link);
#endif
}


/*
 * auth_prealloc_symkeys
 */
static
void
auth_prealloc_symkeys(
        int     keycount
        )
{
        int     allocated;
        int     additional;

        allocated = authnumkeys + authnumfreekeys;
        additional = keycount - allocated;
        if (additional > 0)
                auth_moremem(additional);
        auth_resize_hashtable();
}


static __mbg_inline u_short
auth_log2(double x)
{
        return (u_short)(log10(x) / log10(2));
}


/*
 * auth_resize_hashtable
 *
 * Size hash table to average 4 or fewer entries per bucket initially,
 * within the bounds of at least 4 and no more than 15 bits for the hash
 * table index.  Populate the hash table.
 */
static void
auth_resize_hashtable(void)
{
        u_long          totalkeys;
        u_short         hashbits;
        u_short         hash;
        size_t          newalloc;
        symkey *        sk;

        totalkeys = authnumkeys + authnumfreekeys;
        hashbits = auth_log2(totalkeys / 4.0) + 1;
        hashbits = max(4, hashbits);
        hashbits = min(15, hashbits);

        authhashbuckets = 1 << hashbits;
        authhashmask = authhashbuckets - 1;
        newalloc = authhashbuckets * sizeof(key_hash[0]);

        key_hash = erealloc(key_hash, newalloc);
        memset(key_hash, '\0', newalloc);

        ITER_DLIST_BEGIN(key_listhead, sk, llink, symkey)
                hash = KEYHASH(sk->keyid);
                LINK_SLIST(key_hash[hash], sk, hlink);
        ITER_DLIST_END()
}


/*
 * allocsymkey - common code to allocate and link in symkey
 *
 * secret must be allocated with a free-compatible allocator.  It is
 * owned by the referring symkey structure, and will be free()d by
 * freesymkey().
 */
static void
allocsymkey(
        symkey **       bucket,
        keyid_t         id,
        u_short         flags,
        u_short         type,
        u_long          lifetime,
        u_short         secretsize,
        u_char *        secret
        )
{
        symkey *        sk;

        if (authnumfreekeys < 1)
                auth_moremem(-1);
        UNLINK_HEAD_SLIST(sk, authfreekeys, llink.f);
        DEBUG_ENSURE(sk != NULL);
        sk->keyid = id;
        sk->flags = flags;
        sk->type = type;
        sk->secretsize = secretsize;
        sk->secret = secret;
        sk->lifetime = lifetime;
        LINK_SLIST(*bucket, sk, hlink);
        LINK_TAIL_DLIST(key_listhead, sk, llink);
        authnumfreekeys--;
        authnumkeys++;
}


/*
 * freesymkey - common code to remove a symkey and recycle its entry.
 */
static void
freesymkey(
        symkey *        sk,
        symkey **       bucket
        )
{
        symkey *        unlinked;

        if (sk->secret != NULL) {
                // TODO: why is it set to 0 before it is freed?
                memset(sk->secret, '\0', sk->secretsize);
                free(sk->secret);
        }
        UNLINK_SLIST(unlinked, *bucket, sk, hlink, symkey);
        DEBUG_ENSURE(sk == unlinked);
        UNLINK_DLIST(sk, llink);
        memset((char *)sk + offsetof(symkey, symkey_payload), '\0',
               sizeof(*sk) - offsetof(symkey, symkey_payload));
        LINK_SLIST(authfreekeys, sk, llink.f);
        authnumkeys--;
        authnumfreekeys++;
}


/*
 * auth_findkey - find a key in the hash table
 */
struct savekey *
auth_findkey(
        keyid_t         id
        )
{
        symkey *        sk;

        for (sk = key_hash[KEYHASH(id)]; sk != NULL; sk = sk->hlink) {
                if (id == sk->keyid) {
                        return sk;
                }
        }

        return NULL;
}



#if 0

/*
 * auth_havekey - return TRUE if the key id is zero or known
 */
int
auth_havekey(
        KEY_CACHE *p_cache,
        keyid_t         id
        )
{
        symkey *        sk;

        if (0 == id || p_cache->keyid == id) {
                return TRUE;
        }

        for (sk = key_hash[KEYHASH(id)]; sk != NULL; sk = sk->hlink) {
                if (id == sk->keyid) {
                        return TRUE;
                }
        }

        return FALSE;
}

#endif



/*
 * authhavekey - return TRUE and cache the key, if zero or both known
 *               and trusted.
 */
int
authhavekey(
        KEY_CACHE *p_cache,
        keyid_t         id
        )
{
        symkey *        sk;

        authkeylookups++;
        if (0 == id || p_cache->keyid == id) {
                return TRUE;
        }

        /*
         * Seach the bin for the key. If found and the key type
         * is zero, somebody marked it trusted without specifying
         * a key or key type. In this case consider the key missing.
         */
        authkeyuncached++;
        for (sk = key_hash[KEYHASH(id)]; sk != NULL; sk = sk->hlink) {
                if (id == sk->keyid) {
                        if (0 == sk->type) {
                                authkeynotfound++;
                                return FALSE;
                        }
                        break;
                }
        }

        /*
         * If the key is not found, or if it is found but not trusted,
         * the key is not considered found.
         */
        if (NULL == sk) {
                authkeynotfound++;
                return FALSE;
        }
        if (!(KEY_TRUSTED & sk->flags)) {
                authnokey++;
                return FALSE;
        }

        /*
         * The key is found and trusted. Initialize the key cache.
         * TODO: Check if protection (locking) is required.
         * Instead of duplicating the cached members in 'sk' we could
         * make KEY_CACHE a sub-structure.
         */
        p_cache->keyid = sk->keyid;
        p_cache->type = sk->type;
        p_cache->flags = sk->flags;
        p_cache->secret = sk->secret;
        p_cache->secretsize = sk->secretsize;

        return TRUE;
}


/*
 * authtrust - declare a key to be trusted/untrusted
 */
void
authtrust(
        KEY_CACHE *p_cache,
        keyid_t         id,
        u_long          trust
        )
{
        symkey **       bucket;
        symkey *        sk;
        u_long          lifetime;

        /*
         * Search bin for key; if it does not exist and is untrusted,
         * forget it.
         */
        bucket = &key_hash[KEYHASH(id)];
        for (sk = *bucket; sk != NULL; sk = sk->hlink) {
                if (id == sk->keyid)
                        break;
        }
        if (!trust && NULL == sk)
                return;

        /*
         * There are two conditions remaining. Either it does not
         * exist and is to be trusted or it does exist and is or is
         * not to be trusted.
         */
        if (sk != NULL) {
                if (p_cache->keyid == id) {
                        p_cache->flags = 0;
                        p_cache->keyid = 0;
                }

                /*
                 * Key exists. If it is to be trusted, say so and
                 * update its lifetime.
                 */
                if (trust > 0) {
                        sk->flags |= KEY_TRUSTED;
                        if (trust > 1)
                                sk->lifetime = current_time + trust;
                        else
                                sk->lifetime = 0;
                        return;
                }

                /* No longer trusted, return it to the free list. */
                freesymkey(sk, bucket);
                return;
        }

        /*
         * keyid is not present, but the is to be trusted.  We allocate
         * a new key, but do not specify a key type or secret.
         */
        if (trust > 1) {
                lifetime = current_time + trust;
        } else {
                lifetime = 0;
        }
        allocsymkey(bucket, id, KEY_TRUSTED, 0, lifetime, 0, NULL);
}


/*
 * authistrusted - determine whether a key is trusted
 */
int
authistrusted(
        KEY_CACHE *p_cache,
        keyid_t         keyno
        )
{
        symkey *        sk;
        symkey **       bucket;

        if (keyno == p_cache->keyid)
                return !!(KEY_TRUSTED & p_cache->flags);

        authkeyuncached++;
        bucket = &key_hash[KEYHASH(keyno)];
        for (sk = *bucket; sk != NULL; sk = sk->hlink) {
                if (keyno == sk->keyid)
                        break;
        }
        if (NULL == sk || !(KEY_TRUSTED & sk->flags)) {
                authkeynotfound++;
                return FALSE;
        }
        return TRUE;
}



void
MD5auth_setkey(
        KEY_CACHE *p_cache,
        keyid_t keyno,
        int     keytype,
        const u_char *key,
        int     len
        )
{
        symkey *        sk;
        symkey **       bucket;
        u_char *        secret;
        size_t          secretsize;

        DEBUG_ENSURE(keytype <= USHRT_MAX);
        DEBUG_ENSURE(len < 4 * 1024);
        len = max(0, len);
        /*
         * See if we already have the key.  If so just stick in the
         * new value.
         */
        bucket = &key_hash[KEYHASH(keyno)];
        for (sk = *bucket; sk != NULL; sk = sk->hlink) {
                if (keyno == sk->keyid) {
                        sk->type = (u_short)keytype;
                        secretsize = len;
                        sk->secretsize = (u_short)secretsize;
#ifndef DISABLE_BUG1243_FIX
                        memcpy(sk->secret, key, secretsize);
#else
                        strlcpy((char *)sk->secret, (const char *)key,
                                secretsize);
#endif
                        if (p_cache->keyid == keyno) {
                                p_cache->flags = 0;
                                p_cache->keyid = 0;
                        }
                        return;
                }
        }

        /*
         * Need to allocate new structure.  Do it.
         */
        secretsize = len;
        secret = emalloc(secretsize);
#ifndef DISABLE_BUG1243_FIX
        memcpy(secret, key, secretsize);
#else
        strlcpy((char *)secret, (const char *)key, secretsize);
#endif
        allocsymkey(bucket, keyno, 0, (u_short)keytype, 0,
                    (u_short)secretsize, secret);
#ifdef DEBUG
        if (debug >= 4) {
                size_t  j;

                printf("auth_setkey: key %d type %d len %d ", (int)keyno,
                    keytype, (int)secretsize);
                for (j = 0; j < secretsize; j++)
                        printf("%02x", secret[j]);
                printf("\n");
        }
#endif
}


/*
 * auth_delkeys - delete non-autokey untrusted keys, and clear all info
 *                except the trusted bit of non-autokey trusted keys, in
 *                preparation for rereading the keys file.
 */
void
auth_delkeys(void)
{
        symkey *        sk;

        ITER_DLIST_BEGIN(key_listhead, sk, llink, symkey)
                if (sk->keyid > NTP_MAXKEY) {   /* autokey */
                        continue;
                }

                /*
                 * Don't lose info as to which keys are trusted.
                 */
                if (KEY_TRUSTED & sk->flags) {
                        if (sk->secret != NULL) {
                                memset(sk->secret, '\0', sk->secretsize);
                                free(sk->secret);
                        }
                        sk->secretsize = 0;
                        sk->lifetime = 0;
                } else {
                        freesymkey(sk, &key_hash[KEYHASH(sk->keyid)]);
                }
        ITER_DLIST_END()
}


/*
 * auth_agekeys - delete keys whose lifetimes have expired
 */
void
auth_agekeys(void)
{
        symkey *        sk;

        ITER_DLIST_BEGIN(key_listhead, sk, llink, symkey)
                if (sk->lifetime > 0 && current_time > sk->lifetime) {
                        freesymkey(sk, &key_hash[KEYHASH(sk->keyid)]);
                        authkeyexpired++;
                }
        ITER_DLIST_END()
        DPRINTF(1, ("auth_agekeys: at %lu keys %lu expired %lu\n",
                    current_time, authnumkeys, authkeyexpired));
}


/*
 * authencrypt - generate message authenticator
 *
 * Returns length of authenticator field, zero if key not found.
 */
int
authencrypt(
        KEY_CACHE *p_cache,
        keyid_t         keyno,
        u_int32 *       pkt,
        int             length
        )
{
        /*
         * A zero key identifier means the sender has not verified
         * the last message was correctly authenticated. The MAC
         * consists of a single word with value zero.
         */
        authencryptions++;
        pkt[length / 4] = htonl(keyno);
        if (0 == keyno) {
                return 4;
        }
        if (!authhavekey(p_cache, keyno)) {
                return 0;
        }

        //##++++ TODO: This can be simplified. Check cache usage in MD5authencrypt
        return MD5authencrypt(p_cache, p_cache->type, p_cache->secret, pkt, length);
}


/*
 * authdecrypt - verify message authenticator
 *
 * Returns TRUE if authenticator valid, FALSE if invalid or not found.
 */
int
authdecrypt(
        KEY_CACHE *p_cache,
        keyid_t         keyno,
        u_int32 *       pkt,
        int             length,
        int             size
        )
{
        /*
         * A zero key identifier means the sender has not verified
         * the last message was correctly authenticated.  For our
         * purpose this is an invalid authenticator.
         */
        authdecryptions++;
        if (0 == keyno || !authhavekey(p_cache, keyno) || size < 4) {
                return FALSE;
        }

        return MD5authdecrypt(p_cache, p_cache->type, p_cache->secret, pkt, length,
                              size);
}



#if 1  // from authreadkeys.c

/*
 * authreadkeys.c - routines to support the reading of the key file
 */

#if 0
#include <config.h>
#include <stdio.h>
#include <ctype.h>

#include "ntp_fp.h"
#include "ntp.h"
#include "ntp_syslog.h"
#include "ntp_stdlib.h"

#ifdef OPENSSL
#include "openssl/objects.h"
#include "openssl/evp.h"
#endif  /* OPENSSL */
#endif

/* Forwards */
static char *nexttok (char **);

/*
 * nexttok - basic internal tokenizing routine
 */
static char *
nexttok(
        char    **str
        )
{
        register char *cp;
        char *starttok;

        cp = *str;

        /*
         * Space past white space
         */
        while (*cp == ' ' || *cp == '\t')
                cp++;

        /*
         * Save this and space to end of token
         */
        starttok = cp;
        while (*cp != '\0' && *cp != '\n' && *cp != ' '
               && *cp != '\t' && *cp != '#')
                cp++;

        /*
         * If token length is zero return an error, else set end of
         * token to zero and return start.
         */
        if (starttok == cp)
                return NULL;

        if (*cp == ' ' || *cp == '\t')
                *cp++ = '\0';
        else
                *cp = '\0';

        *str = cp;
        return starttok;
}


/*
 * authreadkeys - (re)read keys from a file.
 */
int
authreadkeys(
        const char *file
        )
{
        KEY_CACHE key_cache = { 0 };
        FILE    *fp;
        char    *line;
        char    *token;
        keyid_t keyno;
        int     keytype;
        char    buf[512];               /* lots of room for line */
        u_char  keystr[32];             /* Bug 2537 */
        int     len;
        int     j;

        /*
         * Open file.  Complain and return if it can't be opened.
         */
        fp = fopen(file, "r");
        if (fp == NULL) {
#if 1  // NOTE: modification by martin
                mbglog( LOG_ERR, "Failed to read file %s: %s",
                        file, strerror( errno ) );
#else
                msyslog(LOG_ERR, "authreadkeys: file %s: %m",
                    file);
#endif
                return (0);
        }
        INIT_SSL();

        /*
         * Remove all existing keys
         */
        auth_delkeys();

        /*
         * Now read lines from the file, looking for key entries
         */
        while ((line = fgets(buf, sizeof buf, fp)) != NULL) {
                token = nexttok(&line);
                if (token == NULL)
                        continue;

                /*
                 * First is key number.  See if it is okay.
                 */
                keyno = atoi(token);
                if (keyno == 0) {
                        msyslog(LOG_ERR,
                            "authreadkeys: cannot change key %s", token);
                        continue;
                }

                if (keyno > NTP_MAXKEY) {
                        msyslog(LOG_ERR,
                            "authreadkeys: key %s > %d reserved for Autokey",
                            token, NTP_MAXKEY);
                        continue;
                }

                /*
                 * Next is keytype. See if that is all right.
                 */
                token = nexttok(&line);
                if (token == NULL) {
                        msyslog(LOG_ERR,
                            "authreadkeys: no key type for key %d", keyno);
                        continue;
                }
#ifdef OPENSSL
                /*
                 * The key type is the NID used by the message digest
                 * algorithm. There are a number of inconsistencies in
                 * the OpenSSL database. We attempt to discover them
                 * here and prevent use of inconsistent data later.
                 */
                keytype = keytype_from_text(token, NULL);
                if (keytype == 0) {
                        msyslog(LOG_ERR,
                            "authreadkeys: invalid type for key %d", keyno);
                        continue;
                }
                if (EVP_get_digestbynid(keytype) == NULL) {
                        msyslog(LOG_ERR,
                            "authreadkeys: no algorithm for key %d", keyno);
                        continue;
                }
#else   /* !OPENSSL follows */

                /*
                 * The key type is unused, but is required to be 'M' or
                 * 'm' for compatibility.
                 */
                if (!(*token == 'M' || *token == 'm')) {
                        msyslog(LOG_ERR,
                            "authreadkeys: invalid type for key %d", keyno);
                        continue;
                }
                keytype = KEY_TYPE_MD5;
#endif  /* !OPENSSL */

                /*
                 * Finally, get key and insert it. If it is longer than 20
                 * characters, it is a binary string encoded in hex;
                 * otherwise, it is a text string of printable ASCII
                 * characters.
                 */
                token = nexttok(&line);
                if (token == NULL) {
                        msyslog(LOG_ERR,
                            "authreadkeys: no key for key %d", keyno);
                        continue;
                }
                len = strlen(token);
                if (len <= 20) {        /* Bug 2537 */
                        MD5auth_setkey(&key_cache, keyno, keytype, (u_char *)token, len);
                } else {
                        char    hex[] = "0123456789abcdef";
                        u_char  temp;
                        char    *ptr;
                        int     jlim;

                        jlim = min(len, 2 * sizeof(keystr));
                        for (j = 0; j < jlim; j++) {
                                ptr = strchr(hex, tolower(token[j]));
                                if (ptr == NULL)
                                        break;  /* abort decoding */
                                temp = (u_char)(ptr - hex);
                                if (j & 1)
                                        keystr[j / 2] |= temp;
                                else
                                        keystr[j / 2] = temp << 4;
                        }
                        if (j < jlim) {
                                msyslog(LOG_ERR,
                                        "authreadkeys: invalid hex digit for key %d", keyno);
                                continue;
                        }
                        MD5auth_setkey(&key_cache, keyno, keytype, keystr, jlim / 2);
                }
        }
        fclose(fp);
        return (1);
}

#endif // from authreadkeys.c




/*HDR*/
/**
 * @brief Check authentication of a received NTP packet
 *
 * @param[in,out]  p_cache  A ::KEY_CACHE structure use to cache keys
 * @param[in]      p_rcv    ::NTP_PACKET_INFO structure containing a received packet
 *
 * @return one of the ::MBG_NTP_AUTH_CODES
 */
int mbg_ntp_auth_decrypt( KEY_CACHE *p_cache, NTP_PACKET_INFO *p_rcv )
{
  keyid_t skeyid;
  int mac_len = get_mac_len( p_rcv );
  int rc;

  if ( mac_len < sizeof( NTP_KEY_ID ) )
    return MBG_NTP_AUTH_NONE;

  skeyid = htonl( p_rcv->packet.mac.key_id );

  if ( skeyid == 0 )   // received a crypto NACK
    return MBG_NTP_AUTH_NACK;

  rc = authdecrypt( p_cache, skeyid, (u_int32 *) &p_rcv->packet,
                    NTP_BASE_PACKET_SIZE, mac_len );

  return ( rc == 0 ) ? MBG_NTP_AUTH_FAIL : MBG_NTP_AUTH_OK;

}  // mbg_ntp_auth_decrypt



/*HDR*/
/**
 * @brief Generate signature for a packet to be sent
 *
 * @param[in,out]  p_cache  A ::KEY_CACHE structure use to cache keys
 * @param[in,out]  p_xmt    ::NTP_PACKET_INFO structure containing the base packet to transmit
 * @param[in]      key_id   A valid key ID, or 0 in case a crypto NACK must be sent
 *
 * @return one of the ::MBG_NTP_AUTH_CODES
 */
int mbg_ntp_auth_encrypt( KEY_CACHE *p_cache, NTP_PACKET_INFO *p_xmt, NTP_KEY_ID key_id )
{
  int auth_len;

  if ( key_id == 0 )  // must send crypto NACK
  {
    p_xmt->packet.mac.key_id = 0;  // no need for htonl() with 0
    p_xmt->len += sizeof( p_xmt->packet.mac.key_id );
    return MBG_NTP_AUTH_NACK;
  }

  auth_len = authencrypt( p_cache, key_id, (uint32_t *) &p_xmt->packet, p_xmt->len );

  p_xmt->len += auth_len;

  if ( auth_len >= sizeof( NTP_KEY_ID ) )
    return MBG_NTP_AUTH_OK;

  return MBG_NTP_AUTH_FAIL;

}  // mbg_ntp_auth_encrypt



/*HDR*/
int mbg_ntp_read_conf_file( const char *fn )
{
  FILE *fp = fopen( fn, "rt" );

  if ( fp == NULL )
  {
    int this_errno = errno;

    mbglog( LOG_WARNING, "Failed to read NTP config file %s: %s",
            fn, strerror( this_errno ) );
    return -this_errno;
  }

  for (;;)
  {
    char s[256];
    char *line = fgets( s, sizeof( s ), fp );
    char *token;

    if ( line == NULL )  // nothing could be read
      break;

    token = nexttok( &line );

    if ( token == NULL )
      continue;

    if ( strcmp( token, "keys" ) == 0 )
    {
      token = nexttok( &line );

      if ( token )
        authreadkeys( token );  //##++ TODO: else error

      continue;
    }

    if ( strcmp( token, "trustedkey" ) == 0 )
    {
      while ( ( token = nexttok( &line ) ) != NULL )
        mbg_ntp_auth_trust_key( atol( token ), 1 );

      continue;
    }

  }  // for (;;)

  fclose( fp );

  return 0;

}  // mbg_ntp_read_conf_file



/*HDR*/
int mbg_ntp_auth_add_key( NTP_KEY_ID key_id, const char *key_type_str, const char *secret )
{
  KEY_CACHE key_cache = { 0 };
  int key_type = keytype_from_text( key_type_str, NULL );  // usually type 'M' is read from the keys file

  if ( key_type == 0 )
  {
    //##++++ error: key type invalid
    return 0;
  }

  // The key type is the NID used by the message digest algorithm.
  // We check if it is supported.
  if ( EVP_get_digestbynid( key_type ) == NULL )
  {
    //##++++ error: key type not supported by openSSL
    return 0;
  }

  // NOTE: the original code in authreadkeys() implementes a special handling
  // if the length of the key hash is > 20. See also NTP bug #2537 for details.
  MD5auth_setkey( &key_cache, key_id, key_type, (u_char *) secret, strlen( secret ) );

  return 1;

}  // mbg_ntp_auth_add_key



/*HDR*/
int mbg_ntp_auth_trust_key( long key_id, int trust )
{
  KEY_CACHE key_cache = { 0 };

  if ( key_id == 0 )  // unspecified
  {
    mbglog( LOG_WARNING, "trusted key ID 0 is invalid" );
    return -1;
  }

  if ( key_id > NTP_MAXKEY )  // reserved for autokey
  {
    mbglog( LOG_WARNING, "trusted key IDs > %i are reserved for autokey", NTP_MAXKEY );
    return -2;
  }

  authtrust( &key_cache, (NTP_KEY_ID) key_id, trust );

  return 0;  // success

}  // mbg_ntp_auth_trust_key



/*HDR*/
/**
 * @brief Initialize SSL and pre-allocate symmetric keys.
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
 int mbg_ntp_auth_init( void )
{
  int rc;

  INIT_SSL();
  rc = init_auth();

  if ( rc < 0 )  // error
    return MBG_ERR_NO_MEM;

  auth_prealloc_symkeys( 0 );  //##++++ TODO or the number of trusted keys we have determined before?

  return MBG_SUCCESS;

}  // mbg_ntp_auth_init



#if DEBUG_MBG_NTP_AUTH

/*HDR*/
int mbg_ntp_auth_test( keyid_t key_id )
{
  // test signature with known packet
  KEY_CACHE key_cache = { 0 };
  uint8_t x[sizeof( NTP_PACKET )] = { 0 };
  NTP_PACKET *p = (NTP_PACKET *) x;
  uint8_t *cp;
  int i;
  int rc;

  int auth_len = authencrypt( &key_cache, key_id, (uint32_t *) x, NTP_BASE_PACKET_SIZE );

  printf( "Test auth: key %i, result len %i\n", key_id, auth_len );

  printf( "packet:" );

  for ( i = 0, cp = x; i <  NTP_BASE_PACKET_SIZE; i++, cp++ )
    printf( " %02X", *cp );

  printf( "\n" );

  printf( "key & hash:" );

  for ( i = 0; i < auth_len; i++, cp++ )
    printf( " %02X", *cp );

  printf( "\n" );

  rc = authdecrypt( &key_cache, key_id, (uint32_t *) p, NTP_BASE_PACKET_SIZE, auth_len );

  return rc;

}  // mbg_ntp_auth_test

#endif  // DEBUG_MBG_NTP_AUTH




