
/**************************************************************************
 *
 *  $Id: mbg_ntp_auth.h 1.4 2019/09/04 07:23:02 martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Definitions and prototypes used with mbg_ntp_auth.c
 *
 * -----------------------------------------------------------------------
 *  $Log: mbg_ntp_auth.h $
 *  Revision 1.4  2019/09/04 07:23:02  martin
 *  mbg_ntp_auth_init() now provides a return code.
 *  Revision 1.3  2018/07/31 14:03:38  martin
 *  Fixed clang compiler warning.
 *  Revision 1.2  2014/09/24 08:56:27  martin
 *  Introduced a key cache structure which can be used to make
 *  the key cache thread-safe.
 *  Updated function prototypes.
 *  Revision 1.1  2014/09/10 15:37:57  martin
 *  Initial revision.
 *
 **************************************************************************/

#ifndef _MBG_NTP_AUTH_H
#define _MBG_NTP_AUTH_H


/* Other headers to be included */

#include <mbgntp.h>

#define OPENSSL  // use openSSL for MD5 directly

#include <openssl/evp.h>
#include <openssl/err.h>


#ifdef _MBG_NTP_AUTH
 #define _ext
 #define _DO_INIT
#else
 #define _ext extern
#endif



#if defined( _USE_PACK )
// not used here
// #pragma pack( 1 )      // set byte alignment
// #define _USING_BYTE_ALIGNMENT
#endif

/* Start of header body */

#ifdef __cplusplus
extern "C" {
#endif

#if !defined( DEBUG_MBG_NTP_AUTH )
  #if defined( DEBUG )
    #define DEBUG_MBG_NTP_AUTH   0  // Set != 0 for debug
  #else
    #define DEBUG_MBG_NTP_AUTH   0  // always 0
  #endif
#endif


#define NTP_COMPAT   1

#if NTP_COMPAT
  typedef uint32_t u_int32;
#endif

typedef uint32_t keyid_t;        /* cryptographic key ID */
#define NTP_MAXKEY      65535    /* maximum symmetric key ID (ntp_crypto.h) */


/* from ntp.h */
#define LEN_PKT_NOMAC   (12 * sizeof(u_int32)) /* min header length */
#define MIN_MAC_LEN     (1 * sizeof(u_int32))   /* crypto_NAK */
#define MAX_MD5_LEN     (5 * sizeof(u_int32))   /* MD5 */
#define MAX_MAC_LEN     (6 * sizeof(u_int32))   /* SHA */



/*
 * The key cache. We cache the last key we looked at here.
 */
#if 1

// A structure holding key cache information.
// Useful to make the auth functions thread-safe.

typedef struct
{
  keyid_t keyid;            /* key identifier */
  u_char *secret;           /* secret */
  u_short secretsize;       /* secret length */
  int     type;             /* OpenSSL digest NID */
  u_short flags;            /* flags that wave */

} KEY_CACHE;

#else

// The original code from the NTP sources
keyid_t cache_keyid;            /* key identifier */
u_char *cache_secret;           /* secret */
u_short cache_secretsize;       /* secret length */
int     cache_type;             /* OpenSSL digest NID */
u_short cache_flags;            /* flags that wave */

#endif






/* function prototypes: */

/* ----- function prototypes begin ----- */

/* This section was generated automatically */
/* by MAKEHDR, do not remove the comments. */

 /**
 * @brief Check authentication of a received NTP packet
 *
 * @param[in,out]  p_cache  A ::KEY_CACHE structure use to cache keys
 * @param[in]      p_rcv    ::NTP_PACKET_INFO structure containing a received packet
 *
 * @return one of the ::MBG_NTP_AUTH_CODES
 */
 int mbg_ntp_auth_decrypt( KEY_CACHE *p_cache, NTP_PACKET_INFO *p_rcv ) ;

 /**
 * @brief Generate signature for a packet to be sent
 *
 * @param[in,out]  p_cache  A ::KEY_CACHE structure use to cache keys
 * @param[in,out]  p_xmt    ::NTP_PACKET_INFO structure containing the base packet to transmit
 * @param[in]      key_id   A valid key ID, or 0 in case a crypto NACK must be sent
 *
 * @return one of the ::MBG_NTP_AUTH_CODES
 */
 int mbg_ntp_auth_encrypt( KEY_CACHE *p_cache, NTP_PACKET_INFO *p_xmt, NTP_KEY_ID key_id ) ;

 int mbg_ntp_read_conf_file( const char *fn ) ;
 int mbg_ntp_auth_add_key( NTP_KEY_ID key_id, const char *key_type_str, const char *secret ) ;
 int mbg_ntp_auth_trust_key( long key_id, int trust ) ;
 int mbg_ntp_auth_init( void ) ;
 int mbg_ntp_auth_test( keyid_t key_id ) ;

/* ----- function prototypes end ----- */

#ifdef __cplusplus
}
#endif


#if defined( _USING_BYTE_ALIGNMENT )
  #pragma pack()      // set default alignment
  #undef _USING_BYTE_ALIGNMENT
#endif

/* End of header body */


#undef _ext
#undef _DO_INIT

#endif  /* _MBG_NTP_AUTH_H */

