
/**************************************************************************
 *
 *  $Id: mbg_ntp_test_util.c 1.11 2018/07/31 16:06:22 martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Functions useful to test NTP implementations.
 *
 * -----------------------------------------------------------------------
 *  $Log: mbg_ntp_test_util.c $
 *  Revision 1.11  2018/07/31 16:06:22  martin
 *  New function ntp_mode_name_str().
 *  Improved printing of packets.
 *  Revision 1.10  2018/07/31 12:12:13  martin
 *  More POSIX compatibility.
 *  Revision 1.9  2018/03/09 09:55:16  martin
 *  Improved printing of refid using functions from new module mbgntp.c.
 *  Moved snprint_ntp_prec() to mbgntp.c.
 *  Revision 1.8  2017/09/06 14:04:26  martin
 *  Root delay and root dispersion are now printed
 *  with 6 instead of only 4 fractional digits.
 *  Revision 1.7  2016/12/01 17:06:50  martin
 *  Account for modified packet structure.
 *  Print leap smear offset decoded from refid field.
 *  Revision 1.6  2016/11/21 17:22:49  martin
 *  Adjustable client poll value sent to server, defaults to 6.
 *  Revision 1.5  2016/08/05 12:35:43  martin
 *  Proper evaluation of precision field.
 *  Use safe str_util functions.
 *  Revision 1.4  2016/07/18 14:53:28Z  martin
 *  Use new functions from module timeutil.
 *  Fixed compiler warning under Windows.
 *  Revision 1.3  2016/07/14 09:17:02  martin
 *  Fixed build under Windows.
 *  Revision 1.2  2014/09/24 09:25:44Z  martin
 *  Account for modified NTP_PACKET definition.
 *  Account for changed NTP_TSTAMP field names.
 *  Fixed some compiler warnings.
 *  Avoid using MINLONG.
 *  Added some functions.
 *  Added and updated doxygen comments.
 *  Revision 1.1  2014/04/01 13:02:21  martin
 *  Initial revision.
 *
 **************************************************************************/

#define _MBG_NTP_TEST_UTIL
  #include <mbg_ntp_test_util.h>
#undef _MBG_NTP_TEST_UTIL

#include <timeutil.h>
#include <str_util.h>
#include <mbgerror.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#if defined( MBG_TGT_LINUX )
  #include <limits.h>      // LONG_MIN, LONG_MAX
#elif defined( MBG_TGT_BSD )
  #include <sys/limits.h>  // LONG_MIN, LONG_MAX
#endif

#include <float.h>         // DBL_MIN, DBL_MAX



/*HDR*/
/**
 * @brief Return a string associated with a packet mode
 *
 * @param[in] mode  The mode number from an NTP packet, see ::NTP_MODES
 *
 * @return  An associated mode name string
 */
const char *ntp_mode_name_str( int mode )
{
  static const char *mode_name_strs[N_NTP_MODES] = NTP_MODE_NAME_STRS;

  return ( mode < N_NTP_MODES ) ? mode_name_strs[mode] : "(unknown)";

};  // ntp_mode_name_str



/*HDR*/
/**
 * @brief Reset query statistics
 *
 * @param[out] p  Statistics structure to be cleared.
 */
void reset_query_stats( QUERY_STATS *p )
{
  memset( p, 0, sizeof( *p ) );

  p->turnaround.min = LONG_MAX;
  p->turnaround.max = LONG_MIN;
  p->latency.min = DBL_MAX;
  p->latency.max = DBL_MIN;
  p->ntp_delay.min = DBL_MAX;
  p->ntp_delay.max = DBL_MIN;
  p->ntp_offs.min = DBL_MAX;
  p->ntp_offs.max = DBL_MIN;

}  // reset_query_stats



/*HDR*/
/**
 * @brief Print a statistics value of type "double"
 *
 * @param[in] info        Optional text string printed at beginning of line, or NULL.
 * @param[in] info_width  Width of the info text field.
 * @param[in] curr        Current statistics parameter.
 * @param[in] curr_width  With of the printed field containing the current parameter value.
 * @param[in] scale       Scale for the parameter values before they are printed.
 * @param[in] p           Optional minimum and maximum values of this parameter, or NULL.
 * @param[in] hint        Optional hint text appended at the end of the line, or NULL.
 */
void print_stats_double( const char *info, int info_width,
                         double curr, int curr_width, double scale,
                         const MIN_MAX_DOUBLE *p, const char *hint )
{
  if ( info )
    printf( "%-*s", info_width, info );

  printf( "%*.3f", curr_width, curr * scale );

  if ( p )
    printf( " (%.3f..%.3f)", p->min * scale, p->max * scale );

  if ( hint )
    printf( "%s", hint );

  printf( "\n" );

}  // print_stats_double



/*HDR*/
/**
 * @brief Print a "struct timespec" as calendar date and time
 *
 * @param[in] p_ts  The time stamp to be printed, in struct timespec format.
 */
void print_timespec_date_time( const struct timespec *p_ts )
{
  struct tm tm = { 0 };
  time_t t = cvt_to_time_t( p_ts->tv_sec );
  int rc = mbg_gmtime( &tm, &t );

  if ( mbg_rc_is_success( rc ) )
  {
    printf( "%04i-%02i-%02i %02i:%02i:%02i.%09li",
            tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
            tm.tm_hour, tm.tm_min, tm.tm_sec,
            (long) p_ts->tv_nsec
          );
  }
  else
  {
    char s[100];
    snprint_gmtime_error( s, sizeof( s ), rc, t, __func__ );
    printf( "%s", s );
  }

}  // print_timespec_date_time



/*HDR*/
/**
 * @brief Print date and time of an NTP time stamp
 *
 * @param[in] t_ntp  The time stamp to be printed, in NTP format.
 */
void print_ntp_date_time( const NTP_TSTAMP *t_ntp )
{
  struct timespec ts;

  ntp_tstamp_to_timespec( &ts, t_ntp );
  print_timespec_date_time( &ts );

}  // print_ntp_date_time



/*HDR*/
/**
 * @brief Print a raw NTP timestamp, and optionally the derived date and time
 *
 * @param[in] msg  Text string printed at the beginning of the line.
 * @param[in] t    Time stamp, in NTP format.
 * @param[in] tail Optional text appended to the output, or NULL.
 * @param[in] print_date_time  If not 0, print also the derived date and time.
 */
void print_ntp_time( const char *msg, const NTP_TSTAMP *t, const char *tail, int print_date_time )
{
  printf( "%s%08lX.%08lX", msg, (ulong) t->seconds, (ulong) t->fractions );

  if ( print_date_time )
  {
    printf( "  " );
    print_ntp_date_time( t );
  }

  if ( tail )
    printf( "%s", tail );

}  // print_ntp_time



/*HDR*/
/**
 * @brief Print contents of an NTP packet on several lines
 *
 * @param[in] msg  Header text to be printed on first line.
 * @param[in] p    An ::NTP_PACKET_INFO structure containing the NTP packet plus some additional information.
 * @param[in] print_date_time  If not 0, print also the date and time derived from the NTP time stamps.
 */
void print_ntp_packet( const char *msg, const NTP_PACKET_INFO *p, int print_date_time, int verbose )
{
  const NTP_BASE_PACKET *pbp = &p->packet.base;
  size_t i;
  char ref_id_str_txt[5];
  char ref_id_str_bin[16];
  int8_t pkt_precision = pbp->flags.precision;
  char prec_str[40];
  bool refid_should_be_text = ntp_refid_should_be_text( pbp->flags.stratum, &pbp->refid );

  snprint_ntp_prec( prec_str, sizeof( prec_str ), pkt_precision );

  printf( "%s: mode %u (%s)\n", msg, pbp->flags.mode,
          ntp_mode_name_str( pbp->flags.mode ) );

  printf( INDENT
    "version %u, leap %u, stratum %u, poll %u, prec %i (%s)\n",
    pbp->flags.version,
    pbp->flags.leap_ind,
    pbp->flags.stratum,
    pbp->flags.poll,
    pkt_precision,
    prec_str
  );

  snprintf_ntp_refid_as_text( ref_id_str_txt, sizeof( ref_id_str_txt ), &pbp->refid );
  snprintf_ntp_refid_as_dotted_quad( ref_id_str_bin, sizeof( ref_id_str_bin ), &pbp->refid );

  printf(
    INDENT "root delay:      %08lX (%.6f s)\n"
    INDENT "root dispersion: %08lX (%.6f s/s)\n"
    INDENT "reference id:    %08lX "
    ,
    (long) pbp->root_delay, _u_fp_to_d( pbp->root_delay ),
    (long) pbp->root_dispersion, _u_fp_to_d( pbp->root_dispersion ),
    (long) pbp->refid.as_u32
  );

  if ( refid_should_be_text )
  {
    printf( "(\"%s\"", ref_id_str_txt );

    if ( verbose )
      printf( ", %s", ref_id_str_bin );
  }
  else  // refid is an IPv4 address or IPv6 address hash
  {
    printf( "(%s", ref_id_str_bin );

    if ( verbose )
      printf( ", \"%s\"", ref_id_str_txt );
  }
  printf( ")\n" );


  print_ntp_time( INDENT "Ref time:       ", &pbp->reference_time, newline, print_date_time );
  print_ntp_time( INDENT "Org time (T1):  ", &pbp->originate_time, newline, print_date_time );
  print_ntp_time( INDENT "Rcv time (T2):  ", &pbp->receive_time, newline, print_date_time );
  print_ntp_time( INDENT "Xmt time (T3):  ", &pbp->transmit_time, newline, print_date_time );


  if ( timespec_is_set( &p->tsr ) )
  {
    // We could also print the timespec directly but actually convert
    // it to NTP format first to have the same output formatting.
    NTP_TSTAMP ts;
    timespec_to_ntp_tstamp( &ts, &p->tsr );
    print_ntp_time( INDENT "Curr time (T4): ", &ts, newline, print_date_time );
  }

  if ( packet_has_mac( p ) )
  {
    const NTP_MAC *pm = &p->packet.mac;
    size_t digest_len = get_digest_len( p );

    printf( INDENT "Auth key ID:    %08X", pm->key_id );

    if ( pm->key_id == 0 && digest_len == 0 )
      printf( " (crypto NACK)" );

    printf( "\n" );

    if ( digest_len > 0 )
    {
      printf( INDENT "Auth digest:    " );

      for ( i = 0; i < digest_len; i++ )
        printf( "%02X", pm->digest[i] );
    }

    printf( "\n" );
  }

  printf( "\n" );

}  // print_ntp_packet



/*HDR*/
/**
 * @brief Initialize a basic NTP packet
 *
 * @param[in] p             The ::NTP_PACKET to be initialized.
 * @param[in] req_mode      Initalizer for the packet's mode field, e.g. ::MODE_CLIENT.
 * @param[in] prot_version  Initalizer for the packet's version field, e.g. ::DEFAULT_REQ_NTP_VERSION.
 */
void init_ntp_req_packet( NTP_PACKET *p, int req_mode, int prot_version )
{
  NTP_BASE_PACKET *pbp = &p->base;

  memset( p, 0, sizeof( *p ) );

  pbp->flags.mode = req_mode;
  pbp->flags.version = prot_version;
  pbp->flags.leap_ind = 3;
  pbp->flags.stratum = 0;
  pbp->flags.poll = client_poll_value;
  pbp->flags.precision = client_precision;

}  // init_ntp_req_packet



/*HDR*/
/**
 * @brief Evaluate the time stamps from an NTP packet received from a server
 *
 * @note The packet must already have been converted to host byte order.
 *
 * @param[out] p_rslt  ::NTP_RESULTS structure to take up the results of the computations.
 * @param[in]  p_rcv   The ::NTP_PACKET_INFO containing the packet received from the server, in host byte order.
 *
 * @see ::eval_ntp_packet
 * @see ::update_query_stats
 */
void get_ntp_results( NTP_RESULTS *p_rslt, const NTP_PACKET_INFO *p_rcv )
{
  const NTP_BASE_PACKET *pbp = &p_rcv->packet.base;

  // T4 is the packet receive time which is not part of
  // the NTP base packet but passed via the NTP_PACKET_INFO
  // structure. Save the timespec converted to NTP format
  // in the result record.
  timespec_to_ntp_tstamp( &p_rslt->t_rcv, &p_rcv->tsr );

  // delay = (T4 - T1) - (T3 - T2)
  p_rslt->turnaround = delta_tstamps( &p_rslt->t_rcv, &pbp->originate_time );  // (T4 - T1), our own clock domain
  p_rslt->latency = delta_tstamps( &pbp->transmit_time, &pbp->receive_time );  // (T3 - T2), server's clock domain
  p_rslt->ntp_delay = p_rslt->turnaround - p_rslt->latency;                    // (T4 - T1) - (T3 - T2)

  // offset = ((T2 - T1) + (T3 - T4)) / 2
  p_rslt->t21 = delta_tstamps( &pbp->receive_time, &pbp->originate_time );     // (T2 - T1)
  p_rslt->t34 = delta_tstamps( &pbp->transmit_time, &p_rslt->t_rcv );          // (T3 - T4)
  p_rslt->ntp_offs = ( p_rslt->t21 + p_rslt->t34 ) / 2.0;                      // ((T2 - T1) + (T3 - T4)) / 2

}  // get_ntp_results



/*HDR*/
/**
 * @brief Update query statistics with the results from the last query
 *
 * @param[in,out] p_stats      A ::QUERY_STATS structure to be updated.
 * @param[in]     p_rslt       ::NTP_RESULTS structure eventually set up by ::get_ntp_results.
 * @param[in]     new_queries  number of queries sent to yield the latest results (usually 1).
 *
 * @see ::get_ntp_results
 */
void update_query_stats( QUERY_STATS *p_stats, const NTP_RESULTS *p_rslt, int new_queries )
{
  p_stats->replies += new_queries;

  update_min_max_double( p_rslt->turnaround, &p_stats->turnaround );
  update_min_max_double( p_rslt->latency, &p_stats->latency );
  update_min_max_double( p_rslt->ntp_offs, &p_stats->ntp_offs );
  update_min_max_double( p_rslt->ntp_delay, &p_stats->ntp_delay );

}  // update_query_stats



/*HDR*/
/**
 * @brief Print query statistics
 *
 * @param[in] p_rslt      ::NTP_RESULTS structure eventually set up by ::get_ntp_results.
 * @param[in] p_stats     Optional ::QUERY_STATS structure, or NULL.
 * @param[in] info_width  Field with of the info strings.
 * @param[in] val_width   Field width for the values.
 * @param[in] scale       Scale for the values.
 */
void print_query_stats( const NTP_RESULTS *p_rslt, const QUERY_STATS *p_stats,
                        int info_width, int val_width, double scale )
{
  print_stats_double( "turnaround: ", info_width, p_rslt->turnaround, val_width, scale,
                      p_stats ? &p_stats->turnaround : NULL, " us  (T4 - T1)" );

  print_stats_double( "server latency: ", info_width, p_rslt->latency, val_width, scale,
                      p_stats ? &p_stats->latency : NULL, " us  (T3 - T2)" );

  print_stats_double( "computed delay: ", info_width, p_rslt->ntp_delay, val_width, scale,
                      p_stats ? &p_stats->ntp_delay : NULL, " us  ((T4 - T1) - (T3 - T2))" );

  print_stats_double( "computed offset: ", info_width, p_rslt->ntp_offs, val_width, scale,
                      p_stats ? &p_stats->ntp_offs : NULL, " us  (((T2 - T1) + (T3 - T4)) / 2)" );

}  // print_query_stats



/*HDR*/
/**
 * @brief Evaluate an NTP packet received from a server
 *
 * @note The packet to be evaluated must be in network byte order
 * and is converted to host byte order.
 *
 * @param[out]     p_rslt   Optional ::NTP_RESULTS structure to take up the results of the computations, or NULL.
 * @param[in,out]  p_rcv    The ::NTP_PACKET to be evaluated, in network byte order.
 * @param[in,out]  p_stats  A ::QUERY_STATS structure to be updated.
 *
 * @see ::ntoh_ntp_packet
 * @see ::get_ntp_results
 * @see ::update_query_stats
 */
int eval_ntp_packet( NTP_RESULTS *p_rslt, NTP_PACKET_INFO *p_rcv,
                     QUERY_STATS *p_stats )
{
  NTP_RESULTS rslt;

  if ( p_rslt == NULL )  // optional parameter not passed
    p_rslt = &rslt;

  // Message authentication code (MAC) should be checked here, if supported.

  ntoh_ntp_packet( p_rcv );

  get_ntp_results( p_rslt, p_rcv );

  if ( p_stats )
    update_query_stats( p_stats, p_rslt, 1 );

  return 1;

}  // eval_ntp_packet



/*HDR*/
/**
 * @brief Eval NTP packet and update statistics without returning results
 *
 * @param[in,out]  p_rcv    The ::NTP_PACKET to be evaluated, in network byte order.
 * @param[in,out]  p_stats  A ::QUERY_STATS structure to be updated.
 *
 * @see ::eval_ntp_packet
 */
int update_ntp_stats( NTP_PACKET_INFO *p_rcv, QUERY_STATS *p_stats )
{
  return eval_ntp_packet( NULL, p_rcv, p_stats );

}  // update_ntp_stats



/*HDR*/
/**
 * @brief Print NTP results
 *
 * @param[in]  p_rslt   ::NTP_RESULTS structure eventually set up by ::eval_ntp_packet, or NULL.
 * @param[in]  p_req    The ::NTP_PACKET_INFO containing the request packet sent to the server, in host byte order, or NULL.
 * @param[in]  p_rcv    The ::NTP_PACKET_INFO containing the packet received from the server, in host byte order, or NULL.
 * @param[in]  p_stats  A ::QUERY_STATS structure to be printed, or NULL.
 * @param[in]  print_date_time  Controls if ::print_ntp_packet also prints broken down date and time
 * @param[in]  verbose          Also prints a line with auth result if auth is not used at all
 *
 * @see ::eval_ntp_packet
 */
void print_ntp_results( const NTP_RESULTS *p_rslt, const NTP_PACKET_INFO *p_req,
                        const NTP_PACKET_INFO *p_rcv, const QUERY_STATS *p_stats,
                        int print_date_time, int verbose )
{
  if ( p_req )
    print_ntp_packet( "Request packet", p_req, print_date_time, verbose );

  if ( p_rcv )
  {
    double d;
    bool smearing_active;
    print_ntp_packet( "Response packet", p_rcv, print_date_time, verbose );
    smearing_active = leap_smearing_active( &p_rcv->packet.base.refid, &d );
    if ( verbose || smearing_active )
    {
      if ( smearing_active )
        printf( "Leap smearing active, offset: %.6f s\n", d );
      else
        printf( "Leap smearing not active\n" );

      printf( "\n" );
    }
  }

  if ( p_rslt )  // p_stats may be NULL
  {
    const char *cp = NULL;

    switch ( p_rslt->auth_code )
    {
      case MBG_NTP_AUTH_NONE: cp = verbose ? "not used" : NULL; break;
      case MBG_NTP_AUTH_OK:   cp = "OK"; break;
      case MBG_NTP_AUTH_NACK: cp = "received crypto NACK";  break;
      case MBG_NTP_AUTH_FAIL: cp = "*** failed ***"; break;
    }

    if ( cp )
      printf( "Authentication: %s\n\n", cp );

    print_query_stats( p_rslt, p_stats, 17, 10, 1e6 );

    printf( "\n" );
  }

}  // print_ntp_results



