
/**************************************************************************
 *
 *  $Id: mbgntp.h 1.14 2018/07/31 16:07:33 martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Definitions and prototypes for mbgntp.h.
 *
 * -----------------------------------------------------------------------
 *  $Log: mbgntp.h $
 *  Revision 1.14  2018/07/31 16:07:33  martin
 *  Changed packet modes to enum.
 *  Added initializer for mode name string array.
 *  Revision 1.13  2018/03/09 09:57:33  martin
 *  New inline function ntp_refid_should_be_text().
 *  Updated function prototypes.
 *  Revision 1.12  2017/12/14 16:25:19  martin
 *  Fixed some clang compiler warnings.
 *  Revision 1.11  2017/09/07 08:11:37  martin
 *  Renamed NTP_FLAGS to NTP_FLAG_FIELD to avoid
 *  a naming collision.
 *  Revision 1.10  2017/08/15 15:45:02  martin
 *  Define NSECS_PER_SEC only if it hasn't been defined before.
 *  Revision 1.9  2016/12/01 17:05:24  martin
 *  Use new union NTP_REFID or the packet's refid field.
 *  New function leap_smearing_active() which also decodes the offset from the refid.
 *  Revision 1.8  2016/07/14 08:17:23  martin
 *  Fixed build under Windows.
 *  Revision 1.7  2014/10/28 08:52:33Z  martin
 *  Made definition of NSECS_PER_SEC the same as in mbgtime.h to
 *  avoid compiler warnings in projects using both header files.
 *  Added/updated some comments.
 *  Revision 1.6  2014/09/24 11:18:39  martin
 *  Modified NTP_PACKET definition which now always includes the base
 *  packet and the maximum sized message authentication code (MAC).
 *  New structure NTP_PACKET_INFO which includes an NTP_PACKET plus
 *  the current packet length, depending on the MAC usage, plus the
 *  transmit or receive time stamp in struct timespec format.
 *  Inline functions can be used to tell if a packet actually includes
 *  a MAC field with or without digest, return its length, extract its
 *  key ID, etc.
 *  Renamed NTP_OFFS to NTP_SEC_BIAS.
 *  Added or modified some doxygen comments.
 *  Reordered some code.
 *  Revision 1.5  2014/05/06 11:03:47  martin
 *  Use NTP_TSTAMP from gpsdefs.h.
 *  Revision 1.4  2014/04/01 12:30:30  martin
 *  Include mbg_tgt.h.
 *  MD5 auth support disabled by default.
 *  Moved some definitions here.
 *  Revision 1.3  2005/11/15 13:51:15  martin
 *  Changed NTP_TSTAMP::fraction to unsigned.
 *  Conditional preliminary union for NTP_TSTAMP using 64 bit type.
 *  Revision 1.2  2005/05/19 14:42:07  martin
 *  Use C99 fixed with data types.
 *  Support MD5 symmetric key authentication.
 *  Revision 1.1  2000/05/23 12:04:21  martin
 *  Initial revision
 *
 **************************************************************************/

#ifndef _MBGNTP_H
#define _MBGNTP_H


/* Other headers to be included */

#include <mbg_tgt.h>
#include <words.h>
#include <gpsdefs.h>

#if defined( MBG_TGT_POSIX )
  #include <arpa/inet.h>  // ntohl()
#endif


#ifdef _MBGNTP
 #define _ext
 #define _DO_INIT
#else
 #define _ext extern
#endif


/* Start of header body */


/**
 * @brief The IANA-assigned network port used by NTP
 */
#define NTP_PORT 123



/**
 * @brief NTP epoch bias from ordinary time_t epoch
 *
 * The Unix time_t epoch is usually 1970-01-01 00:00 whereas
 * the NTP epoch is 1900-01-01 00:00, so the difference is
 * a constant number of seconds:<br>
 *
 * time_t t = ntp_time - ::NTP_SEC_BIAS
 */
#define NTP_SEC_BIAS      2208988800UL


#if !defined( NSECS_PER_SEC )
  #define NSECS_PER_SEC     1000000000UL
#endif


#define NTP_FRAC_PER_SEC         (uint64_t) 4294967296.0

#define MIN_REQ_NTP_VERSION      3
#define MAX_REQ_NTP_VERSION      4
#define DEFAULT_REQ_NTP_VERSION  MAX_REQ_NTP_VERSION


enum NTP_MODES
{
  MODE_UNSPEC,    // 0, unspecified (old version)
  MODE_ACTIVE,    // 1, symmetric active mode
  MODE_PASSIVE,   // 2, symmetric passive mode
  MODE_CLIENT,    // 3, client mode
  MODE_SERVER,    // 4, server mode
  MODE_BROADCAST, // 5, broadcast mode
  // Modes 6 and 7 also exist, but aren't yet supported here.
  N_NTP_MODES     // number of defined modes
};


#define NTP_MODE_NAME_STRS      \
{                               \
  "unspecified (old version)",  \
  "symmetric active",           \
  "symmetric passive",          \
  "client",                     \
  "server",                     \
  "broadcast"                   \
}



/**
 * @brief Codes returned by functions generating or checking NTP authentication
 */
enum MBG_NTP_AUTH_CODES
{
  MBG_NTP_AUTH_NONE,    ///< Auth. not used or supported
  MBG_NTP_AUTH_OK,      ///< Auth. checking on receive was OK, try auth on send
  MBG_NTP_AUTH_NACK,    ///< Received a crypto NACK (i.e. key ID == 0)
  MBG_NTP_AUTH_FAIL,    ///< Auth. checking on receive failed, must send crypto NACK
  N_MBG_NTP_AUTH_CODES  ///< Number of known codes
};



// types and conversion functions stolen from NTP

typedef int32_t NTP_S_FP;   // signed floating point in integer representation
typedef uint32_t NTP_U_FP;  // unsigned floating point in integer representation


/*
 * s_fp/double and u_fp/double conversions
 */

#define FRIC        65536.0  ///< 2^16 as double


#if 1  // use inline functions for conversion

static __mbg_inline /*HDR*/
NTP_S_FP d_to_s_fp( double d )
{
  return (NTP_S_FP) ( d * FRIC );
}

static __mbg_inline /*HDR*/
NTP_S_FP d_to_u_fp( double d )
{
  return (NTP_U_FP) ( d * FRIC );
}

static __mbg_inline /*HDR*/
double s_fp_to_d( NTP_S_FP s_fp )
{
  return (double) s_fp / FRIC;
}

static __mbg_inline /*HDR*/
double u_fp_to_d( NTP_U_FP u_fp )
{
  return (double) u_fp / FRIC;
}

#define _d_to_s_fp( _r )  d_to_s_fp( _r )
#define _d_to_u_fp( _r )  d_to_u_fp( _r )
#define _s_fp_to_d( _r )  s_fp_to_d( _r )
#define _u_fp_to_d( _r )  u_fp_to_d( _r )

#else  // use conversion macros as defined in the NTP source code

#define _d_to_s_fp( _r )  ( (NTP_S_FP) ( (_r) * FRIC ) )
#define _d_to_u_fp( _r )  ( (NTP_U_FP) ( (_r) * FRIC ) )
#define _s_fp_to_d( _r )  _u_fp_to_d( _r )
#define _u_fp_to_d( _r )  ( (double) ( _r ) / FRIC )

#endif



/**
 * @brief The flags field of an ::NTP_BASE_PACKET
 *
 * This is a bit-coded 32 bit wide field.
 */
typedef struct
{
  unsigned int mode      : 3;
  unsigned int version   : 3;  ///< see ::MIN_REQ_NTP_VERSION, ::MAX_REQ_NTP_VERSION, ::DEFAULT_REQ_NTP_VERSION
  unsigned int leap_ind  : 2;
  unsigned int stratum   : 8;
  unsigned int poll      : 8;
  unsigned int precision : 8;

} NTP_FLAG_FIELD;



/**
 * @brief The refid field used in NTP packets
 *
 * This field can be interpreted in different ways.
 * It can contain a 4 character ASCII string <b>without terminating 0</b>,
 * or an IPv4 address, or the hash of an IPv6 address. In case of an NTP
 * server where leap second smearing is in progress the contents is a
 * pseudo IPv4 address 254.x.y.z where x.y.z have to be interpreted as
 * signed 24 bit number, scaled to the range +/-2, representing the
 * current leap smear offset.
 * Endianess needs to be taken into account.
 *
 * @see ::leap_smear_offset_from_refid
 * @see https://support.ntp.org/bin/view/Dev/LeapSmearForNTPv4
 * @see http://bugs.ntp.org/show_bug.cgi?id=2855
 */
typedef union
{
  uint32_t as_u32;
  uint8_t as_u8s[sizeof( uint32_t )];
  char as_chars[sizeof( uint32_t )];

} NTP_REFID;



/**
 * @brief The NTP base packet
 *
 * If authentication is used an ::NTP_MAC field can be appended to the packet,
 * and the length of the MAC field depends on the type of authentication.
 * Also, the packet doesn't include a time stamp when the packet has been received,
 * so the ::NTP_PACKET_INFO structure has been introduced which contains the base
 * packet, plus its actual length, plus a receive or transmit time stamp, and thus
 * simplifies handling of the packet by the implementation.
 *
 * @see ::NTP_MAC
 * @see ::NTP_PACKET
 * @see ::NTP_PACKET_INFO
 */
typedef struct
{
  NTP_FLAG_FIELD flags;
  NTP_U_FP root_delay;
  NTP_U_FP root_dispersion;
  NTP_REFID refid;
  NTP_TSTAMP reference_time;
  NTP_TSTAMP originate_time;
  NTP_TSTAMP receive_time;
  NTP_TSTAMP transmit_time;

} NTP_BASE_PACKET;

#define NTP_BASE_PACKET_SIZE  sizeof( NTP_BASE_PACKET )


/**
 * @brief The key ID field of an ::NTP_MAC
 */
typedef uint32_t NTP_KEY_ID;



/**
 * @brief The maximum length of a message digest (hash)
 *
 * The number of bytes actually used depends on the type of the key
 * (MD5, etc.) the ID of which has been specified in the ::NTP_KEY_ID.
 */
#define MAX_NTP_DIGEST_BYTES   20



/**
 * @brief NTP Message authentication code
 */
typedef struct
{
  /// The ID of the key to be used to generate the ::NTP_MAC::digest.
  NTP_KEY_ID key_id;

  /// The message digest, usually 16 or 20 bytes, depending on key type.
  uint8_t digest[MAX_NTP_DIGEST_BYTES];

} NTP_MAC;



/**
 * @brief An NTP packet including maximum sized MAC
 */
typedef struct
{
  NTP_BASE_PACKET base;   ///< The base NTP packet
  NTP_MAC mac;            ///< The variable length MAC, depending on used authentication

} NTP_PACKET;



/**
 * @brief An NTP packet including additional information
 */
typedef struct
{
  size_t len;            ///< The actual length of the ::NTP_PACKET_INFO::packet
  struct timespec tsr;   ///< A time stamp (CLOCK_REALTIME) when packet has been sent or received
  NTP_PACKET packet;     ///< the packet to be sent, or which has been received

} NTP_PACKET_INFO;




/**
 * @brief Return the length of an NTP packet's MAC
 *
 * @param[in]  p  The ::NTP_PACKET_INFO containing the packet and the actual packet length
 *
 * @return The actual length of the packet's ::NTP_MAC field
 */
static __mbg_inline /*HDR*/
size_t get_mac_len( const NTP_PACKET_INFO *p )
{
  ssize_t mac_len = p->len - NTP_BASE_PACKET_SIZE;

  // don't return negative values
  return ( mac_len < 0 ) ? 0 : mac_len;

}  // get_mac_len



/**
 * @brief Return the length of an NTP packet's message digest
 *
 * @param[in]  p  The ::NTP_PACKET_INFO containing the packet and the actual packet length
 *
 * @return The actual length of the packet's message digest
 */
static __mbg_inline /*HDR*/
size_t get_digest_len( const NTP_PACKET_INFO *p )
{
  ssize_t digest_len = get_mac_len( p ) - sizeof( NTP_KEY_ID );

  // don't return negative values
  return ( digest_len < 0 ) ? 0 : digest_len;

}  // get_mac_len



/**
 * @brief Check if an NTP packet has a MAC
 *
 * @param[in]  p  The ::NTP_PACKET_INFO containing the packet and the actual packet length
 *
 * @return != 0 if the packet has a MAC, else 0
 */
static __mbg_inline /*HDR*/
int packet_has_mac( const NTP_PACKET_INFO *p )
{
  return get_mac_len( p ) >= sizeof( NTP_KEY_ID );

}  // packet_has_mac



/**
 * @brief Check if an NTP packet has a message digest
 *
 * @param[in]  p  The ::NTP_PACKET_INFO containing the packet and the actual packet length
 *
 * @return != 0 if the packet has a digest, else 0
 */
static __mbg_inline /*HDR*/
int packet_has_digest( const NTP_PACKET_INFO *p )
{
  return get_mac_len( p ) > sizeof( NTP_KEY_ID );

}  // packet_has_digest



/**
 * @brief Convert an ::NTP_TSTAMP's byte order from net to host
 *
 * @param[in,out]  t  The time stamp top be converted
 */
static __mbg_inline /*HDR*/
void ntoh_ntp_time( NTP_TSTAMP *t )
{
  t->seconds = ntohl( t->seconds );
  t->fractions = ntohl( t->fractions );

}  // ntoh_ntp_time



/**
 * @brief Convert an ::NTP_PACKET's byte order from net to host
 *
 * @param[in,out]  p  The ::NTP_PACKET_INFO containing the packet to be converted, and the actual packet length
 */
static __mbg_inline /*HDR*/
void ntoh_ntp_packet( NTP_PACKET_INFO *p )
{
  NTP_BASE_PACKET *pbp = &p->packet.base;

  ntoh_ntp_time( &pbp->reference_time );
  ntoh_ntp_time( &pbp->originate_time );
  ntoh_ntp_time( &pbp->receive_time );
  ntoh_ntp_time( &pbp->transmit_time );

  pbp->root_delay = ntohl( pbp->root_delay );
  pbp->root_dispersion = ntohl( pbp->root_dispersion );

  if ( packet_has_mac( p ) )
  {
    p->packet.mac.key_id = ntohl( p->packet.mac.key_id );
    // No need to convert the digest part, if available.
  }

}  // ntoh_ntp_packet



/**
 * @brief Compute the difference of two time stamps in ::NTP_TSTAMP format
 *
 * @param[in]  t      The original time stamp
 * @param[in]  t_ref  The time stamp to be subtracted
 *
 * @return The time difference, in seconds
 */
static __mbg_inline /*HDR*/
double delta_tstamps( const NTP_TSTAMP *t,
                      const NTP_TSTAMP *t_ref )
{
  return ( (double) t->seconds - (double) t_ref->seconds )
       + ( (double) t->fractions - (double) t_ref->fractions ) / NTP_FRAC_PER_SEC;

}  // delta_tstamps



/**
 * @brief Check refid if leap smearing active, and decode smear offset
 *
 * @param[in]   p       The refid field received in an NTP packet from a server
 * @param[out]  p_offs  Used to return the decoded leap smear offset, if applicable, else 0
 *
 * @return true if leap smearing is active, else false.
 */
static __mbg_inline /*HDR*/
bool leap_smearing_active( const NTP_REFID *p, double *p_offs )
{
  // If smearing is active the refid contains a pseudo IP address
  // 254.x.y.z where x.y.z contains the current smear offset.
  // See ::NTP_REFID
  bool smearing_active = p->as_u8s[0] == 254;

  if ( p_offs )  // we need to return the current offset
  {
    double offs = 0.0;  // default, if smearing not active

    if ( smearing_active )
    {
      int32_t tmp = ( ntohl( p->as_u32 ) << 8 ) & 0xFFFFFF00UL;
      offs = ( (double) tmp ) / (double) 0x40000000;
    }

    *p_offs = offs;
  }

  return smearing_active;

}  // leap_smearing_active



static __mbg_inline /*HDR*/
/**
 * @brief Check if the refid in an NTP packet should be ASCII text or binary
 *
 * Depending on the stratum number, the refid can be interpreted
 * as ASCII string up to 4 characters, or as a binary number representing
 * an IPv4 address or the 4 LSBs of the MD5 hash of an IPv6 address.
 *
 * @param[out]  p   Pointer to an ::NTP_BASE_PACKET
 *
 * @return true if the refid should be ASCII, else false.
 *
 * @see ::snprintf_ntp_refid_as_text
 * @see ::snprintf_ntp_refid_as_dotted_quad
 */
bool ntp_refid_should_be_text( int stratum, const NTP_REFID *p )
{
  return ( stratum <= 1 || stratum >= 16 )
         && !leap_smearing_active( p, NULL );

}  // ntp_refid_should_be_text



/* function prototypes: */

#ifdef __cplusplus
extern "C" {
#endif

/* ----- function prototypes begin ----- */

/* This section was generated automatically */
/* by MAKEHDR, do not remove the comments. */

 /**
 * @brief Print the value of the 'precision' field from an NTP packet to a string buffer
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in] prec_exp    The precision value from the NTP packet.
 */
 int snprint_ntp_prec( char *s, size_t max_len, int8_t prec_exp ) ;

 /**
 * @brief Print an NTP refid as text to a string buffer
 *
 * The function ::ntp_refid_should_be_text can be called to determine
 * if the refid is to be interpreted as printable text, so this function
 * can be used to print it into a string buffer.
 *
 * Please note the 4 char refid may not be 0-terminated, so you can't
 * simply cast the original refid to a string, and then print it.
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in]   p         Pointer to a ::NTP_REFID.
 *
 * @return Number of characters copied to the destination buffer, without the terminating 0
 *
 * @see ::ntp_refid_should_be_text
 * @see ::snprintf_ntp_refid_as_dotted_quad
 */
 int snprintf_ntp_refid_as_text( char *s, size_t max_len, const NTP_REFID *p ) ;

 /**
 * @brief Print an NTP refid as dotted quad to a string buffer
 *
 * The function ::ntp_refid_should_be_text can be called to determine
 * if the refid is to be interpreted as printable text, in which case
 * ::snprintf_ntp_refid_as_text should be used to print it, or as binary
 * number, in which case this function should be called to print it.
 *
 * Historically, a binary refid was always the IPv4 address of the upstream
 * time source, so it is usually displayed as dotted quad string.
 * Actually the upstream IP address can also be an IPv6 address which
 * doesn't fit into the 4 byte refid field, so in case of an IPv6 address
 * the binary refid are the 4 LSBs of the MD5 hash of the IPv6 address.
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in]   p         Pointer to a ::NTP_REFID.
 *
 * @return Number of characters copied to the destination buffer, without the terminating 0
 *
 * @see ::ntp_refid_should_be_text
 * @see ::snprintf_ntp_refid_as_dotted_quad
 */
 int snprintf_ntp_refid_as_dotted_quad( char *s, size_t max_len, const NTP_REFID *p ) ;


/* ----- function prototypes end ----- */

#ifdef __cplusplus
}
#endif


/* End of header body */

#undef _ext
#undef _DO_INIT

#endif  /* _MBGNTP_H */

